#!/usr/bin/env node
'use strict';

// Dependencies
const _ = require('lodash');
const $ = require('highland');
const util = require('util');
const yargs = require('yargs');
const files = require('../lib/files');
const functions = require('../lib/functions');
const errors = require('../lib/errors');
const logging = require('../lib/logging');
const Action = require('../lib/actions');
const program = require('../lib/program');
const Promise = require('bluebird');
const microtime = require('../lib/microtime');
const Table = require('easy-table');
const prettyJson = require('prettyjson');

// Logger
const logger = logging.createLogger('nodus-run');

function isStream(value) {
    return util.isFunction(value.pipe);
}

function print_result(value, options = {}) {

    if (util.isNullOrUndefined(value))
        return; // Print Nothing

    if (isStream(value)) {
        return new Promise((resolve, reject) =>
            $(value)
                .each(result => print_result(result))
                .stopOnError(err => reject(err))
                .done(() => resolve())
        );
    }

    // Write a string as is to the console
    if (util.isString(value)) {
        return process.stdout.write(value + '\n');
    }

    // Write the results as a JSON string
    if (util.isObject(value)) {
        logger.debug("FORMAT_OPTIONS", options);

        if (options["table"])
        {
            logger.debug("FORMAT_TABLE");
            return print_result(Table.print(value));
        }

        // TODO: Add indent option to CLI args
        if (options["json"])
            return (!!options["indent"] || options["no-indent"] || process.stdout.isTTY === false)
                ? print_result(JSON.stringify(value))
                : print_result(JSON.stringify(value, null, 2));

        if (options["inspect"])
            return print_result(util.inspect(value, {colors: true}));

        if (options["pretty-json"])
            return print_result(prettyJson.render(value));

        // If no format options were specified then:
        // - use INSPECT if TTY is true
        // - use JSON otherwise (likely piping data)
        return (process.stdout.isTTY)
            // ? print_result(util.inspect(value, {colors: true}))
            ? print_result(prettyJson.render(value))
            : print_result(JSON.stringify(value)) // NO NEW LINE BY DEFAULT
    }

    // Write the string representation of the value
    process.stdout.write(value.toString() + '\n');
}

function print_error(err) {
    console.error(err);
}

/**
 * Load a command to run from the target
 * @param target
 * @returns {*}
 */
function loadAction(targetFile, argv) {
    // Load the module/file to be run
    const target = files.requireFile(targetFile);

    const name = target.name || targetFile;
    logger.debug('TARGET_NAME', name);

    // Load Action
    if (util.isFunction(target)) {
        logger.debug("TARGET_FUNCTION", "Target is a function...");
        return Action(name, target);
    }

    if (util.isObject(target)) {
        logger.debug("TARGET_OBJECT", "Target is an object...");

        const cmd = argv._.shift();
        if (!cmd) {
            logger.debug("No command specified...");
            logger.debug("Printing list of actions...");

            const actions = _.keys(target)
                .filter(prop => util.isFunction(target[prop]));

            // Print Actions and Exit
            // TODO: Prompt user
            logger.warn("**** Please specify a command ****");
            print_result(util.inspect(target, {colors: true}));
            process.exit();
        }

        logger.debug("Loading command...", cmd);
        const command = target[cmd];
        if (util.isNullOrUndefined(command))
        // TODO: print list of available commands
            throw errors("COMMAND_NOT_FOUND", "The command specified was not found on the target.", cmd);

        return Action(`name:${cmd}`, command);
    }

    // Target must be a string or some other unsupported value
    logger.warn("Unknown target type");
    return target;
}

// Program Help
yargs
    .usage('Usage: $0 <command> [args] [options]')
    .boolean('print-args')
    .describe('print-args', 'Print argument list passed to function without invoking it.')
    .default('print-args', false)
    .epilog('Copyright (c) 2017 Brad Serbu <bradserbu@bradserbu.com>\nAll rights reserved.')
    .version()
    .help();

// Parse CommandLine Arguments
const argv = yargs.argv;
logger.debug('ARGV', argv);

// Determine the file/module to be run
const targetFile = argv._.shift();
logger.debug('TARGET_FILE', targetFile);

// Display help if no target is specified
if (util.isNullOrUndefined(targetFile)) {
    return yargs.showHelp();
}

// Load command to run
const action = loadAction(targetFile, argv);

// If we failed to load an action, then we are done...
if (!action)
    return;

const isNamedArg = arg => {
    logger.debug('IS_NAMED_ARG', arg);
    arg = arg.toString();
    return arg && (arg.startsWith("-") === false) && (arg.indexOf('=') !== -1);
};

const namedArgPair = arg => [
    arg.substr(0, arg.indexOf('=')),
    arg.substr(arg.indexOf("=") + 1)
];

// **** PARSE COMMAND LINE ARGUMENTS
// TODO: Support passing function parameters
logger.debug('ARGS', argv._);
const namedArgs = _.fromPairs(argv._.filter(isNamedArg).map(namedArgPair));
logger.debug("NAMED_ARGS", namedArgs);

// TODO: We should be building an argument list, not mapping named args
const args = _.reject(argv._, isNamedArg);

const options = _.omit(argv, ['_', '$0']);
logger.debug("OPTIONS", options);

// ** Print ARGUMENTS
if (options["print-args"]) {
    // PRINT ARGS -> EXIT
    return print_result({
        argv,
        options,
        namedArgs,
        args
    }, options);
}

// Log Events
const started_on = microtime.now();
action.on('completed', data => {

    const elapsed_ms = ((microtime.now() - started_on) * .001).toFixed(2);

    logger.debug(`[COMPLETED]`, data, `(${elapsed_ms} ms)`);
    // logger.debug(`[${event.toUpperCase()}]`, data, `(${elapsed_ms} ms)`)
});

// Map named parameters
const parameters = action.parameters;
logger.debug('PARAMETERS', parameters);

// Consume an argument
const takeArg = () => args.shift();

// List of arguments to pass to the action
const argList = parameters.map(param => {

    // Use named argument first
    if (namedArgs.hasOwnProperty(param))
        return namedArgs[param];

    // Consume an "unnamed/positional" argument
    return takeArg();
});

// Run the activity and display the result
action
    // .mapNamedArguments()
    .run(...argList) // TODO: Build argument list from combination of named args and command ...args
    .then(result => {

        // ** "run" something after invoking the command
        if (options["run"]) {

            const prop = options["run"];
            if (util.isNullOrUndefined(result[prop])) {
                throw errors("NOT_FOUND", "The 'run' action was not found.", {
                    action: prop,
                    result: result
                });
            }

            // ** Invoke the action
            //logger.debug('PARAMETERS', functions.getParameterNames(result[prop]));
            return Action(prop, result[prop])
                .run(...args); // TODO: Do not forward on arguments that were used in the primary action
        }

        // TODO: Apply other options (i.e. formatting, conditionchecks, select property, etc.)

        // Nothing else to do now
        return result;
    })
    .then(result => print_result(result, options))
    .catch(err => print_error(err));

